#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern unsigned long get_cycle(void);
extern void *memcpy_v1(void *, void *, size_t len);
extern void *memcpy_v2(void *, void *, size_t len);
extern void *memcpy_v3(void *, void *, size_t len);
extern void *memcpy_c_v1(void *, void *, size_t len);
extern void *memcpy_c_v2(void *, void *, size_t len);

int compare(const void* a, const void* b)
{
    return (*(unsigned long*)a - *(unsigned long*)b);
}

void print_result(unsigned long result[])
{
    int i;
    unsigned long avg = 0;
    
    qsort(result, 20, sizeof(unsigned long), compare);
    printf("Min: %lu\n", result[0]);
    printf("Max: %lu\n", result[19]);
    
    for (i = 0; i < 20; i++) {
        avg += result[i];
    }
    
    printf("Avg: %lu\n", avg / 20);
}

int main(int argc, char *argv[])
{
    char *buf1[8192];
    char *buf2[8192];
    unsigned long result[20];
    int i;
    unsigned long tsc_start, tsc_end;
    
    memset(buf1, 0xFF, 8192);
    
    printf("glibc (system) memcpy():\n");
    
    for (i = 0; i < 20; i++) {
        tsc_start = get_cycle();
        memcpy(buf2, buf1, 8192);
        tsc_end = get_cycle();
        result[i] = tsc_end - tsc_start;
    }
    
    print_result(result);
    
    printf("C: memcpy_c_v1():\n");
    
    for (i = 0; i < 20; i++) {
        tsc_start = get_cycle();
        memcpy_c_v1(buf2, buf1, 8192);
        tsc_end = get_cycle();
        result[i] = tsc_end - tsc_start;
    }
    
    print_result(result);
    
    printf("C: memcpy_c_v2():\n");
    
    for (i = 0; i < 20; i++) {
        tsc_start = get_cycle();
        memcpy_c_v2(buf2, buf1, 8192);
        tsc_end = get_cycle();
        result[i] = tsc_end - tsc_start;
    }
    
    print_result(result);
    
    printf("Assembly: memcpy_v1():\n");
    
    for (i = 0; i < 20; i++) {
        tsc_start = get_cycle();
        memcpy_v1(buf2, buf1, 8192);
        tsc_end = get_cycle();
        result[i] = tsc_end - tsc_start;
    }
    
    print_result(result);
    
    printf("Assembly: memcpy_v2():\n");
    
    for (i = 0; i < 20; i++) {
        tsc_start = get_cycle();
        memcpy_v2(buf2, buf1, 8192);
        tsc_end = get_cycle();
        result[i] = tsc_end - tsc_start;
    }
    
    print_result(result);
    
    printf("Assembly (with SSE2): memcpy_v3():\n");
    
    for (i = 0; i < 20; i++) {
        tsc_start = get_cycle();
        memcpy_v3(buf2, buf1, 8192);
        tsc_end = get_cycle();
        result[i] = tsc_end - tsc_start;
    }
    
    print_result(result);
    
    return 0;
}
